 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Memory management
  *
  * (c) 1995 Bernd Schmidt
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "uae.h"
#include "memory.h"
#include "ersatz.h"
#include "zfile.h"
#include "newcpu.h"

/* Set by each memory handler that does not simply access real memory.  */
int special_mem;

int ersatzkickfile = 0;

uae_u32 allocated_chipmem;
uae_u32 allocated_fastmem;
uae_u32 allocated_bogomem;
uae_u32 allocated_gfxmem;
uae_u32 allocated_z3fastmem;
uae_u32 allocated_a3000mem;

#ifdef SAVE_MEMORY_BANKS
addrbank *mem_banks[65536];
#else
addrbank mem_banks[65536];
#endif

uae_u8 *baseaddr[65536];

/* This has two functions. It either holds a
 * host address that, when added to the 68k
 * address, gives the host address corresponding
 * to that 68k address (in which case the value
 * in this array is even), OR it holds the same
 * value
 * as mem_banks, for those banks that have
 * baseaddr==0. In that case, bit 0 is set (the
 * memory access routines will take care of it).
 */

#ifdef NO_INLINE_MEMORY_ACCESS
__inline__ uae_u32
longget(uaecptr addr)
{
	return call_mem_get_func(get_mem_bank(addr).lget, addr);
}

__inline__ uae_u32
wordget(uaecptr addr)
{
	return call_mem_get_func(get_mem_bank(addr).wget, addr);
}

__inline__ uae_u32
byteget(uaecptr addr)
{
	return call_mem_get_func(get_mem_bank(addr).bget, addr);
}
__inline__ void
longput(uaecptr addr, uae_u32 l)
{
	call_mem_put_func(get_mem_bank(addr).lput, addr, l);
}
__inline__ void
wordput(uaecptr addr, uae_u32 w)
{
	call_mem_put_func(get_mem_bank(addr).wput, addr, w);
}
__inline__ void
byteput(uaecptr addr, uae_u32 b)
{
	call_mem_put_func(get_mem_bank(addr).bput, addr, b);
}
#endif

uae_u32 chipmem_mask, kickmem_mask, bogomem_mask, a3000mem_mask;

static void
go_debugger(void)
{
	leave_graphics_mode(1);
	activate_debugger();
}

static int
valid_addr(uaecptr addr)
{
	return addr == 0x00200000
		|| (addr >= 0x00F00000 && addr <  0x00F80000)
		|| (addr >= 0x07F00000 && addr <= 0x07F00004)
		|| (addr >= 0x08000000 && addr <= 0x08000004)
		|| (addr >= 0xFF000000 && addr <  0xFF010000);
}

/* A dummy bank that only contains zeros */

static uae_u32 dummy_lget(uaecptr) REGPARAM;
static uae_u32 dummy_wget(uaecptr) REGPARAM;
static uae_u32 dummy_bget(uaecptr) REGPARAM;
static void dummy_lput(uaecptr, uae_u32) REGPARAM;
static void dummy_wput(uaecptr, uae_u32) REGPARAM;
static void dummy_bput(uaecptr, uae_u32) REGPARAM;
static int dummy_check(uaecptr addr, uae_u32 size) REGPARAM;

uae_u32 REGPARAM2
dummy_lget(uaecptr addr)
{
	special_mem |= S_READ;
	if (currprefs.illegal_mem)
	{
		if (!valid_addr(addr))
		{
			write_log("Illegal lget at %08lx\n", addr);
			go_debugger();
		}
	}

	return 0;
}

uae_u32 REGPARAM2
dummy_wget(uaecptr addr)
{
	special_mem |= S_READ;
	if (currprefs.illegal_mem)
	{
		if (!valid_addr(addr))
		{
			write_log("Illegal wget at %08lx\n", addr);
			go_debugger();
		}
	}

	return 0;
}

uae_u32 REGPARAM2
dummy_bget(uaecptr addr)
{
	special_mem |= S_READ;
	if (currprefs.illegal_mem)
	{
		if (!valid_addr(addr))
		{
			write_log("Illegal bget at %08lx\n", addr);
			go_debugger();
		}
	}

	return 0;
}

void REGPARAM2
dummy_lput(uaecptr addr, uae_u32 l)
{
	special_mem |= S_WRITE;
	if (currprefs.illegal_mem)
	{
		if (!valid_addr(addr))
		{
			write_log("Illegal lput at %08lx\n", addr);
			go_debugger();
		}
	}
}

void REGPARAM2
dummy_wput(uaecptr addr, uae_u32 w)
{
	special_mem |= S_WRITE;
	if (currprefs.illegal_mem)
	{
		if (!valid_addr(addr))
		{
			write_log("Illegal wput at %08lx\n", addr);
			go_debugger();
		}
	}
}

void REGPARAM2
dummy_bput(uaecptr addr, uae_u32 b)
{
	special_mem |= S_WRITE;
	if (currprefs.illegal_mem)
	{
		if (!valid_addr(addr))
		{
			write_log("Illegal bput at %08lx\n", addr);
			go_debugger();
		}
	}
}

int REGPARAM2
dummy_check(uaecptr addr, uae_u32 size)
{
	special_mem |= S_READ;
	if (currprefs.illegal_mem)
	{
		if (!valid_addr(addr))
		{
			write_log("Illegal check at %08lx\n", addr);
			go_debugger();
		}
	}
	return 0;
}

/* A3000 "motherboard resources" bank.  */
static uae_u32
mbres_lget(uaecptr)
	REGPARAM;
	static uae_u32 mbres_wget(uaecptr) REGPARAM;
	static uae_u32 mbres_bget(uaecptr) REGPARAM;
	static void mbres_lput(uaecptr, uae_u32) REGPARAM;
	static void mbres_wput(uaecptr, uae_u32) REGPARAM;
	static void mbres_bput(uaecptr, uae_u32) REGPARAM;
	static int mbres_check(uaecptr addr, uae_u32 size) REGPARAM;

	static int mbres_val = 0;

	uae_u32 REGPARAM2 mbres_lget(uaecptr addr)
{
	special_mem |= S_READ;
	if (currprefs.illegal_mem)
	{
		write_log("A3000 motherboard: Illegal lget at %08lx\n", addr);
		go_debugger();
	}

	return 0;
}

uae_u32 REGPARAM2
mbres_wget(uaecptr addr)
{
	special_mem |= S_READ;
	if (currprefs.illegal_mem)
	{
		write_log("A3000 motherboard: Illegal wget at %08lx\n", addr);
		go_debugger();
	}

	return 0;
}

uae_u32 REGPARAM2
mbres_bget(uaecptr addr)
{
	special_mem |= S_READ;
	if (currprefs.illegal_mem)
	{
		write_log("A3000 motherboard: Illegal bget at %08lx\n", addr);
		go_debugger();
	}

	return (addr & 0xFFFF) == 3 ? mbres_val : 0;
}

void REGPARAM2
mbres_lput(uaecptr addr, uae_u32 l)
{
	special_mem |= S_WRITE;
	if (currprefs.illegal_mem)
	{
		write_log("A3000 motherboard: Illegal lput at %08lx\n", addr);
		go_debugger();
	}
}

void REGPARAM2
mbres_wput(uaecptr addr, uae_u32 w)
{
	special_mem |= S_WRITE;
	if (currprefs.illegal_mem)
	{
		write_log("A3000 motherboard: Illegal wput at %08lx\n", addr);
		go_debugger();
	}
}

void REGPARAM2
mbres_bput(uaecptr addr, uae_u32 b)
{
	special_mem |= S_WRITE;
	if (currprefs.illegal_mem)
	{
		write_log("A3000 motherboard: Illegal bput at %08lx\n", addr);
		go_debugger();
	}

	if ((addr & 0xFFFF) == 3)
		mbres_val = b;
}

int REGPARAM2
mbres_check(uaecptr addr, uae_u32 size)
{
	if (currprefs.illegal_mem)
	{
		write_log("A3000 motherboard: Illegal check at %08lx\n", addr);
		go_debugger();
	}

	return 0;
}

/* Chip memory */

uae_u8 *chipmemory;

static int
chipmem_check(uaecptr addr, uae_u32 size)
	REGPARAM;
	static uae_u8 *chipmem_xlate(uaecptr addr) REGPARAM;

	uae_u32 REGPARAM2 chipmem_lget(uaecptr addr)
{
	uae_u32 *m;

	addr -= chipmem_start & chipmem_mask;
	addr &= chipmem_mask;
	m = (uae_u32 *) (chipmemory + addr);
	return do_get_mem_long(m);
}

uae_u32 REGPARAM2
chipmem_wget(uaecptr addr)
{
	uae_u16 *m;

	addr -= chipmem_start & chipmem_mask;
	addr &= chipmem_mask;
	m = (uae_u16 *) (chipmemory + addr);
	return do_get_mem_word(m);
}

uae_u32 REGPARAM2
chipmem_bget(uaecptr addr)
{
	addr -= chipmem_start & chipmem_mask;
	addr &= chipmem_mask;
	return chipmemory[addr];
}

void REGPARAM2
chipmem_lput(uaecptr addr, uae_u32 l)
{
	uae_u32 *m;

	addr -= chipmem_start & chipmem_mask;
	addr &= chipmem_mask;
	m = (uae_u32 *) (chipmemory + addr);
	do_put_mem_long(m, l);
}

void REGPARAM2
chipmem_wput(uaecptr addr, uae_u32 w)
{
	uae_u16 *m;

	addr -= chipmem_start & chipmem_mask;
	addr &= chipmem_mask;
	m = (uae_u16 *) (chipmemory + addr);
	do_put_mem_word(m, w);
}

void REGPARAM2
chipmem_bput(uaecptr addr, uae_u32 b)
{
	addr -= chipmem_start & chipmem_mask;
	addr &= chipmem_mask;
	chipmemory[addr] = b;
}

int REGPARAM2
chipmem_check(uaecptr addr, uae_u32 size)
{
	addr -= chipmem_start & chipmem_mask;
	addr &= chipmem_mask;
	return (addr + size) <= allocated_chipmem;
}

uae_u8 REGPARAM2 *
chipmem_xlate(uaecptr addr)
{
	addr -= chipmem_start & chipmem_mask;
	addr &= chipmem_mask;
	return chipmemory + addr;
}

/* Slow memory */

static uae_u8 *bogomemory;

static uae_u32
bogomem_lget(uaecptr)
	REGPARAM;
	static uae_u32 bogomem_wget(uaecptr) REGPARAM;
	static uae_u32 bogomem_bget(uaecptr) REGPARAM;
	static void bogomem_lput(uaecptr, uae_u32) REGPARAM;
	static void bogomem_wput(uaecptr, uae_u32) REGPARAM;
	static void bogomem_bput(uaecptr, uae_u32) REGPARAM;
	static int bogomem_check(uaecptr addr, uae_u32 size) REGPARAM;
	static uae_u8 *bogomem_xlate(uaecptr addr) REGPARAM;

	uae_u32 REGPARAM2 bogomem_lget(uaecptr addr)
{
	uae_u32 *m;

	addr -= bogomem_start & bogomem_mask;
	addr &= bogomem_mask;
	m = (uae_u32 *) (bogomemory + addr);
	return do_get_mem_long(m);
}

uae_u32 REGPARAM2
bogomem_wget(uaecptr addr)
{
	uae_u16 *m;

	addr -= bogomem_start & bogomem_mask;
	addr &= bogomem_mask;
	m = (uae_u16 *) (bogomemory + addr);
	return do_get_mem_word(m);
}

uae_u32 REGPARAM2
bogomem_bget(uaecptr addr)
{
	addr -= bogomem_start & bogomem_mask;
	addr &= bogomem_mask;
	return bogomemory[addr];
}

void REGPARAM2
bogomem_lput(uaecptr addr, uae_u32 l)
{
	uae_u32 *m;

	addr -= bogomem_start & bogomem_mask;
	addr &= bogomem_mask;
	m = (uae_u32 *) (bogomemory + addr);
	do_put_mem_long(m, l);
}

void REGPARAM2
bogomem_wput(uaecptr addr, uae_u32 w)
{
	uae_u16 *m;

	addr -= bogomem_start & bogomem_mask;
	addr &= bogomem_mask;
	m = (uae_u16 *) (bogomemory + addr);
	do_put_mem_word(m, w);
}

void REGPARAM2
bogomem_bput(uaecptr addr, uae_u32 b)
{
	addr -= bogomem_start & bogomem_mask;
	addr &= bogomem_mask;
	bogomemory[addr] = b;
}

int REGPARAM2
bogomem_check(uaecptr addr, uae_u32 size)
{
	addr -= bogomem_start & bogomem_mask;
	addr &= bogomem_mask;
	return (addr + size) <= allocated_bogomem;
}

uae_u8 REGPARAM2 *
bogomem_xlate(uaecptr addr)
{
	addr -= bogomem_start & bogomem_mask;
	addr &= bogomem_mask;
	return bogomemory + addr;
}

/* A3000 motherboard fast memory */

static uae_u8 *a3000memory;

static uae_u32
a3000mem_lget(uaecptr)
	REGPARAM;
	static uae_u32 a3000mem_wget(uaecptr) REGPARAM;
	static uae_u32 a3000mem_bget(uaecptr) REGPARAM;
	static void a3000mem_lput(uaecptr, uae_u32) REGPARAM;
	static void a3000mem_wput(uaecptr, uae_u32) REGPARAM;
	static void a3000mem_bput(uaecptr, uae_u32) REGPARAM;
	static int a3000mem_check(uaecptr addr, uae_u32 size) REGPARAM;
	static uae_u8 *a3000mem_xlate(uaecptr addr) REGPARAM;

	uae_u32 REGPARAM2 a3000mem_lget(uaecptr addr)
{
	uae_u32 *m;

	addr -= a3000mem_start & a3000mem_mask;
	addr &= a3000mem_mask;
	m = (uae_u32 *) (a3000memory + addr);
	return do_get_mem_long(m);
}

uae_u32 REGPARAM2
a3000mem_wget(uaecptr addr)
{
	uae_u16 *m;

	addr -= a3000mem_start & a3000mem_mask;
	addr &= a3000mem_mask;
	m = (uae_u16 *) (a3000memory + addr);
	return do_get_mem_word(m);
}

uae_u32 REGPARAM2
a3000mem_bget(uaecptr addr)
{
	addr -= a3000mem_start & a3000mem_mask;
	addr &= a3000mem_mask;
	return a3000memory[addr];
}

void REGPARAM2
a3000mem_lput(uaecptr addr, uae_u32 l)
{
	uae_u32 *m;

	addr -= a3000mem_start & a3000mem_mask;
	addr &= a3000mem_mask;
	m = (uae_u32 *) (a3000memory + addr);
	do_put_mem_long(m, l);
}

void REGPARAM2
a3000mem_wput(uaecptr addr, uae_u32 w)
{
	uae_u16 *m;

	addr -= a3000mem_start & a3000mem_mask;
	addr &= a3000mem_mask;
	m = (uae_u16 *) (a3000memory + addr);
	do_put_mem_word(m, w);
}

void REGPARAM2
a3000mem_bput(uaecptr addr, uae_u32 b)
{
	addr -= a3000mem_start & a3000mem_mask;
	addr &= a3000mem_mask;
	a3000memory[addr] = b;
}

int REGPARAM2
a3000mem_check(uaecptr addr, uae_u32 size)
{
	addr -= a3000mem_start & a3000mem_mask;
	addr &= a3000mem_mask;
	return (addr + size) <= allocated_a3000mem;
}

uae_u8 REGPARAM2 *
a3000mem_xlate(uaecptr addr)
{
	addr -= a3000mem_start & a3000mem_mask;
	addr &= a3000mem_mask;
	return a3000memory + addr;
}

/* Kick memory */

uae_u8 *kickmemory;

static uae_u32
kickmem_lget(uaecptr)
	REGPARAM;
	static uae_u32 kickmem_wget(uaecptr) REGPARAM;
	static uae_u32 kickmem_bget(uaecptr) REGPARAM;
	static void kickmem_lput(uaecptr, uae_u32) REGPARAM;
	static void kickmem_wput(uaecptr, uae_u32) REGPARAM;
	static void kickmem_bput(uaecptr, uae_u32) REGPARAM;
	static int kickmem_check(uaecptr addr, uae_u32 size) REGPARAM;
	static uae_u8 *kickmem_xlate(uaecptr addr) REGPARAM;

	uae_u32 REGPARAM2 kickmem_lget(uaecptr addr)
{
	uae_u32 *m;

	addr -= kickmem_start & kickmem_mask;
	addr &= kickmem_mask;
	m = (uae_u32 *) (kickmemory + addr);
	return do_get_mem_long(m);
}

uae_u32 REGPARAM2
kickmem_wget(uaecptr addr)
{
	uae_u16 *m;

	addr -= kickmem_start & kickmem_mask;
	addr &= kickmem_mask;
	m = (uae_u16 *) (kickmemory + addr);
	return do_get_mem_word(m);
}

uae_u32 REGPARAM2
kickmem_bget(uaecptr addr)
{
	addr -= kickmem_start & kickmem_mask;
	addr &= kickmem_mask;
	return kickmemory[addr];
}

void REGPARAM2
kickmem_lput(uaecptr addr, uae_u32 b)
{
	if (currprefs.illegal_mem)
	{
		write_log("Kick memory: Illegal kickmem lput at %08lx\n", addr);
		go_debugger();
	}
}

void REGPARAM2
kickmem_wput(uaecptr addr, uae_u32 b)
{
	if (currprefs.illegal_mem)
	{
		write_log("Kick memory: Illegal kickmem wput at %08lx\n", addr);
		go_debugger();
	}
}

void REGPARAM2
kickmem_bput(uaecptr addr, uae_u32 b)
{
	if (currprefs.illegal_mem)
	{
		write_log("Kick memory: Illegal kickmem lput at %08lx\n", addr);
		go_debugger();
	}
}

int REGPARAM2
kickmem_check(uaecptr addr, uae_u32 size)
{
	addr -= kickmem_start & kickmem_mask;
	addr &= kickmem_mask;
	return (addr + size) <= kickmem_size;
}

uae_u8 REGPARAM2 *
kickmem_xlate(uaecptr addr)
{
	addr -= kickmem_start & kickmem_mask;
	addr &= kickmem_mask;
	return kickmemory + addr;
}

/* Default memory access functions */

int REGPARAM2
default_check(uaecptr a, uae_u32 b)
{
	return 0;
}

uae_u8 REGPARAM2 *
default_xlate(uaecptr a)
{
	write_log("Your Amiga program just did something terribly stupid at %08x\n", a);
	go_debugger();
	//uae_reset();
	return kickmem_xlate(get_long(0xF80000));	/* So we don't crash. */
}

static int
load_kickstart(void)
{
	int i;
	uae_u32 cksum = 0, prevck = 0;
	unsigned char buffer[20];

	FILE *f = zfile_open(currprefs.romfile, "rb");

	if (f == NULL)
	{
		system("/usr/amiga/t4k");
		f = zfile_open(currprefs.romfile, "rb");
	}
	if (f == NULL)
	{
		write_log("No Kickstart ROM found.\n");
		return 0;
	}

	i = fread(buffer, 1, 11, f);
	if (strncmp((char *) buffer, "AMIROMTYPE1", 11) != 0)
		fseek(f, 0, SEEK_SET);

	i = fread(kickmemory, 1, kickmem_size, f);
	if (i == kickmem_size / 2)
	{
		memcpy(kickmemory + kickmem_size / 2, kickmemory, kickmem_size / 2);
	}
	else if (i != kickmem_size)
	{
		write_log("Error while reading Kickstart.\n");
		return 0;
	}
	zfile_close(f);

	cloanto_rom = 0;
	if (strncmp((char *) buffer, "AMIROMTYPE1", 11) == 0)
	{
		FILE *keyf;
		uae_u8 *p;
		long size, cnt, t;

		cloanto_rom = 1;
		write_log("ROM image is an encrypted \"Amiga Forever\" ROM file.\n");
		if (strlen(currprefs.keyfile) == 0)
		{
			write_log("No filename given for ROM key file.\n");
			return 0;
		}
		keyf = zfile_open(currprefs.keyfile, "rb");
		if (keyf == 0)
		{
			write_log("Could not find ROM key file.\n");
			return 0;
		}
		p = (uae_u8 *) xmalloc(524288);
		size = fread(p, 1, 524288, keyf);

		for (t = cnt = 0; cnt < kickmem_size; cnt++, t = (t + 1) % size)
		{
			kickmemory[cnt] ^= p[t];
			if (i == cnt + 1)
				t = size - 1;
		}
		fclose(keyf);
		free(p);
	}
#if defined(AMIGA)
 chk_sum:
#endif

	for (i = 0; i < kickmem_size; i += 4)
	{
		uae_u32 data =
			kickmemory[i] * 65536 * 256 + kickmemory[i + 1] * 65536 + kickmemory[i +
			2] * 256 + kickmemory[i + 3];
		cksum += data;
		if (cksum < prevck)
			cksum++;
		prevck = cksum;
	}
	if (cksum != 0xFFFFFFFFul)
	{
		write_log
			("Kickstart checksum incorrect. You probably have a corrupted ROM image.\n");
	}
	return 1;
}

/* Address banks */

addrbank dummy_bank = {
	dummy_lget, dummy_wget, dummy_bget,
	dummy_lput, dummy_wput, dummy_bput,
	default_xlate, dummy_check, NULL
};

addrbank mbres_bank = {
	mbres_lget, mbres_wget, mbres_bget,
	mbres_lput, mbres_wput, mbres_bput,
	default_xlate, mbres_check, NULL
};

addrbank chipmem_bank = {
	chipmem_lget, chipmem_wget, chipmem_bget,
	chipmem_lput, chipmem_wput, chipmem_bput,
	chipmem_xlate, chipmem_check, NULL
};

addrbank bogomem_bank = {
	bogomem_lget, bogomem_wget, bogomem_bget,
	bogomem_lput, bogomem_wput, bogomem_bput,
	bogomem_xlate, bogomem_check, NULL
};

addrbank a3000mem_bank = {
	a3000mem_lget, a3000mem_wget, a3000mem_bget,
	a3000mem_lput, a3000mem_wput, a3000mem_bput,
	a3000mem_xlate, a3000mem_check, NULL
};

addrbank kickmem_bank = {
	kickmem_lget, kickmem_wget, kickmem_bget,
	kickmem_lput, kickmem_wput, kickmem_bput,
	kickmem_xlate, kickmem_check, NULL
};

char *address_space, *good_address_map;
int good_address_fd;

static void
init_mem_banks(void)
{
	int i;

	for (i = 0; i < 65536; i++)
		put_mem_bank(i << 16, &dummy_bank, 0);
}

#ifdef NATMEM_OFFSET
shmpiece *shm_start = NULL;
int canbang = 1;

#ifdef NATMEM_MMAP
void *
mapped_malloc(size_t s, char *file)
{
	void *answer;
	int fd;
	shmpiece *x;
	char namebuf[NAME_MAX + 2];

	if (!canbang)
		return malloc(s);

	strcpy(namebuf, "/uae_");
	strcpy(&namebuf[5], file);
	shm_unlink(namebuf);
	fd = shm_open(namebuf, O_RDWR | O_CREAT, 0777);
	if (fd == -1)
	{
		canbang = 0;
		return malloc(s);
	}
	if (ftruncate(fd, s) == -1)
	{
		close(fd);
		shm_unlink(namebuf);
		canbang = 0;
		return malloc(s);
	}
	answer = mmap(NULL, s, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (answer == MAP_FAILED)
	{
		close(fd);
		shm_unlink(namebuf);
		canbang = 0;
		return malloc(s);
	}

	shm_unlink(namebuf);

	x = malloc(sizeof(shmpiece));
	x->native_address = answer;
	x->id = fd;
	x->size = s;
	x->next = shm_start;
	x->prev = NULL;
	if (x->next)
		x->next->prev = x;
	shm_start = x;

	return answer;
}

/* some memory is allocated a different way:
 *
 * it is not remappable but allocated at its final address space. But it is
 * possible to be swapped out (virtual memory). Use this for z3 memory and
 * other fast memory banks. Do not (!) use it for kickstart and chip memory
 * because this has to be remapped at boot time.
 */
void *
mapped_malloc_vmem(size_t s, char *file, uae_u8 * address)
{
	void *answer;
	shmpiece *x;

 retry:
	/* note: MAP_LAZY is somewhat critical. At least under QNX it is not
	 * garuanteed that enough memory is available and uae may be aborted
	 * if more memory is used.
	 *
	 * If MAP_LAZY is not used mmap() should be called once for each MB of
	 * memory or you may not have enough free physical memory free. The number
	 * of mapped "chunks" (each one MB in size) should then be stored in
	 * shmpiece to unmap it correctly later.
	 */
	answer = mmap((void *) address, s,
		PROT_READ | PROT_WRITE, MAP_ANON | MAP_FIXED | MAP_LAZY, NOFD, 0);
	if (answer == MAP_FAILED)
	{
		if (errno == EAGAIN)
		{
			errno = 0;
			goto retry;
		}
		return NULL;
	}

	x = malloc(sizeof(shmpiece));
	x->native_address = answer;
	x->id = NOFD;
	x->size = s;
	x->next = shm_start;
	x->prev = NULL;
	if (x->next)
		x->next->prev = x;
	shm_start = x;

	return answer;
}

#endif

static void
dumplist(void)
{
	shmpiece *x = shm_start;

	printf("Start Dump:\n");
	while (x)
	{
		printf("  this=%p, Native %p, id %d, prev=%p, next=%p, size=0x%08x\n",
			x, x->native_address, x->id, x->prev, x->next, x->size);
		x = x->next;
	}
	printf("End Dump:\n");
}

static void
delete_shmmaps(uae_u32 start, uae_u32 size)
{
	if (!canbang)
		return;

	/*printf("before delete_shmmpas: "); dumplist(); */

	while (size)
	{
		uae_u8 *base = mem_banks[bankindex(start)]->baseaddr;

		if (base)
		{
			shmpiece *x = shm_start;

			base = ((uae_u8 *) NATMEM_OFFSET) + start;
			while (x && x->native_address != base)
				x = x->next;
			if (!x)
			{
#if 0 /* not a serious failure, ignore for now */
				printf("NATMEM: Failure to find mapping at %08x(%p)\n", start, base);
				dumplist();
				canbang = 0;
#endif
				return;
			}
			if (x->size > size)
			{
#if 0 /* not a serious failure, ignore for now */
				printf("NATMEM: Failure to delete mapping at %08x(size %08x, delsize %08x)\n",
					start, x->size, size);
				dumplist();
				canbang = 0;
#endif
				return;
			}
#ifdef NATMEM_MMAP
			if (x->id != NOFD)
				munmap(x->native_address, x->size);
#else
			shmdt(x->native_address);
#endif
			size -= x->size;
			start += x->size;
			if (x->next)
				x->next->prev = x->prev;	/* remove this one from the list */
			if (x->prev)
				x->prev->next = x->next;
			else
				shm_start = x->next;
			free(x);
		}
		else
		{
			size -= 0x10000;
			start += 0x10000;
		}
	}
	/* printf("behind delete_shmmpas: "); dumplist(); */
}

static void
add_shmmaps(uae_u32 start, addrbank * what)
{
	shmpiece *x = shm_start;
	shmpiece *y;
	uae_u8 *base = what->baseaddr;

	if (!canbang)
		return;
	if (!base)
		return;

	while (x && x->native_address != base)
		x = x->next;
	if (!x)
	{
		printf("NATMEM: Failure to find existing at %08x(%p)\n", start, base);
		dumplist();
		canbang = 0;
		return;
	}
	y = malloc(sizeof(shmpiece));
	*y = *x;
	base = ((uae_u8 *) NATMEM_OFFSET) + start;
#ifdef NATMEM_MMAP
	if (y->id != NOFD)
	{
		y->native_address = mmap(base, y->size,
			PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, y->id, 0);
	}
	if (y->native_address == MAP_FAILED)
	{
#else
	y->native_address = shmat(y->id, base, 0);
	if (y->native_address == (void *) -1)
	{
#endif
		printf("NATMEM: Failure to map existing at %08x(%p)\n", start, base);
		perror("shmat");
		dumplist();
		canbang = 0;
		return;
	}
	y->next = shm_start;
	y->prev = NULL;
	if (y->next)
		y->next->prev = y;
	shm_start = y;
	/*  printf("add_shmmaps: "); dumplist(); */
}
#endif

int
memory_init(void)
{
	char buffer[4096];
	char *nam;
	int i, fd;
	int custom_start;
	int ret = 1;

	allocated_chipmem = currprefs.chipmem_size;
	allocated_bogomem = currprefs.bogomem_size;
	allocated_a3000mem = currprefs.a3000mem_size;

	kickmemory = (uae_u8 *) mapped_malloc(kickmem_size, "kick");
	kickmem_bank.baseaddr = (uae_u8 *) kickmemory;

	chipmemory = (uae_u8 *) mapped_malloc(allocated_chipmem, "chip");

	while (!chipmemory && allocated_chipmem > 512 * 1024)
	{
		allocated_chipmem >>= 1;
		chipmemory = (uae_u8 *) mapped_malloc(allocated_chipmem, "chip");
		if (chipmemory)
			write_log("Reducing chipmem size to %dkb\n", allocated_chipmem >> 10);
	}
	if (!chipmemory)
	{
		write_log("virtual memory exhausted (chipmemory)!\n");
		abort();
	}
	chipmem_bank.baseaddr = (uae_u8 *) chipmemory;

	do_put_mem_long((uae_u32 *) (chipmemory + 4), 0);
	init_mem_banks();

	/* Map the chipmem into all of the lower 16MB */
	/* That is a BAD idea if we want to do address space mapping. And
	 * the first 2M, or the total size, whichever is greater,
	 * should really be enough! */
	i = ((allocated_chipmem >> 16) > 32) ? (allocated_chipmem >> 16) : 32;
	map_banks(&chipmem_bank, 0x00, i, allocated_chipmem);
	custom_start = 0xC0;

	map_banks(&custom_bank, custom_start, 0xE0 - custom_start, 0);
	map_banks(&cia_bank, 0xA0, 32, 0);
	map_banks(&clock_bank, 0xDC, 1, 0);
	map_banks(&bonus_bank, 0xDE, 1, 0);

	/* @@@ Does anyone have a clue what should be in the 0x200000 - 0xA00000
	 * range on an Amiga without expansion memory?  */
	custom_start = allocated_chipmem >> 16;
	if (custom_start < 0x20)
		custom_start = 0x20;
	map_banks(&dummy_bank, custom_start, 0xA0 - custom_start, 0);

	if (allocated_bogomem > 0)
	{
		bogomemory = (uae_u8 *) mapped_malloc(allocated_bogomem, "bogo");
		bogomem_bank.baseaddr = (uae_u8 *) bogomemory;
	}
	if (bogomemory != 0)
		map_banks(&bogomem_bank, 0xC0, allocated_bogomem >> 16, allocated_bogomem);
	else
		allocated_bogomem = 0;

	if (allocated_a3000mem > 0)
	{
		a3000memory = (uae_u8 *) mapped_malloc(allocated_a3000mem, "a3000");
		a3000mem_bank.baseaddr = (uae_u8 *) a3000memory;
	}
	if (a3000memory != 0)
		map_banks(&a3000mem_bank, a3000mem_start >> 16,
			allocated_a3000mem >> 16, allocated_a3000mem);
	else
		allocated_a3000mem = 0;

	map_banks(&rtarea_bank, 0xF0, 1, 0);
	if (!load_kickstart())
	{
		ret = 0;
	}
	map_banks(&kickmem_bank, 0xF8, 8, 0);
	map_banks(&expamem_bank, 0xE8, 1, 0);

	if (cloanto_rom)
		map_banks(&kickmem_bank, 0xE0, 8, 0);

	chipmem_mask = allocated_chipmem - 1;
	kickmem_mask = kickmem_size - 1;
	bogomem_mask = allocated_bogomem - 1;
	a3000mem_mask = allocated_a3000mem - 1;
	return ret;
}

void
map_banks(addrbank * bank, int start, int size, int realsize)
{
	int bnr;
	unsigned long int hioffs = 0, endhioffs = 0x100;
	int real_left;
	addrbank *orgbank = bank;
	uae_u32 realstart = start;

	flush_icache();	/* Sure don't want to keep any old mappings around! */
#ifdef NATMEM_OFFSET
	delete_shmmaps(start << 16, size << 16);
#endif

	if (!realsize)
		realsize = size << 16;
	real_left = 0;

	if ((size << 16) < realsize)
	{
		fprintf(stderr, "Please report to bmeyer@cs.monash.edu.au, and mention:\n");
		fprintf(stderr, "Broken mapping, size=%x, realsize=%x\n", size, realsize);
		fprintf(stderr, "Start is %x\n", start);
		fprintf(stderr, "Reducing memory sizes, especially chipmem, may fix this problem\n");
		abort();
	}
	if (start >= 0x100)
	{
		for (bnr = start; bnr < start + size; bnr++)
		{
			if (!real_left)
			{
				realstart = bnr;
				real_left = realsize >> 16;
#ifdef NATMEM_OFFSET
				add_shmmaps(realstart << 16, bank);
#endif
			}
			put_mem_bank(bnr << 16, bank, realstart << 16);
			real_left--;
		}
		return;
	}

	/* Some '020 Kickstarts apparently require a 24 bit address space... */
	if (currprefs.address_space_24)
		endhioffs = 0x10000;
	for (hioffs = 0; hioffs < endhioffs; hioffs += 0x100)
	{
		real_left = 0;
		for (bnr = start; bnr < start + size; bnr++)
		{
			if (!real_left)
			{
				realstart = bnr + hioffs;
				real_left = realsize >> 16;
#ifdef NATMEM_OFFSET
				add_shmmaps(realstart << 16, bank);
#endif
			}
			put_mem_bank((bnr + hioffs) << 16, bank, realstart << 16);
			real_left--;
		}
	}
}

void
unmap_banks(addrbank * bank, int start, int size, int realsize)
{
#ifdef NATMEM_OFFSET
	flush_icache();	/* Sure don't want to keep any old mappings around! */
	delete_shmmaps(start << 16, size << 16);
#endif
}
